/*
 * Ratiometric Frequency Counter, LED version
 *
 * arith.S - arithmetic functions
 *
 * straight forward implementation of long unsigned integer arithmetics
 * optimized for size rather than for speed (see also: AVR200, AVR204)
 *
 * division, multiplication-by-10 and binary-to-bcd routines rely on
 * certain characteristics of input data (see inline comments)
 *
 * routines clobber registers r18-r27, r30, r31, following
 * http://www.nongnu.org/avr-libc/user-manual/FAQ.html#faq_reg_usage
 *
 * (c) Axel (XL) Schwenke, axel.schwenke@gmx.net
 *
 * $Id: arith.S 122 2011-02-21 20:42:55Z schwenke $
 */


/* void mymul(void *prod, void *fac)
 *
 * prod[0..5] = prod[0..2] * fac[0..2]
 *
 */

.global mymul
.func mymul

arg1=           24      /* gcc puts argument1 in r24/25 */
arg2=           22      /* gcc puts argument2 in r22/23 */
loop=           26      /* loop counter */
zreg=           30      /* Z pointer */
zero=           1       /* gcc zero reg */

prod0=          16      /* 6 bytes product */
prod1=          17
prod2=          18
prod3=          19
prod4=          20
prod5=          21

fac0=           27      /* 3 bytes factor */
fac1=           28
fac2=           29


mymul:
                /* save work registers */
                push prod0
                push prod1
                push fac1
                push fac2

                /* load 24 bits from arg2 */
                movw zreg, arg2
                ld fac0, z+
                ld fac1, z+
                ld fac2, z+

                /* load 24 bits from arg1 */
                movw zreg, arg1
                ld prod0, z+
                ld prod1, z+
                ld prod2, z+
                clr prod3
                clr prod4
                sub prod5, prod5

                ldi loop, 25
1:              ror prod5
                ror prod4
                ror prod3
                ror prod2
                ror prod1
                ror prod0
                dec loop
                breq 2f
                brcc 1b
                add prod3, fac0
                adc prod4, fac1
                adc prod5, fac2
                rjmp 1b

                /* store back 48 bit result */
2:              movw zreg, arg1
                st z+, prod0
                st z+, prod1
                st z+, prod2
                st z+, prod3
                st z+, prod4
                st z+, prod5

                /* restore work registers */
                pop fac2
                pop fac1
                pop prod1
                pop prod0

                ret
.endfunc



/* void mydiv(void *prod, void *fac)
 *
 * prod[0..3] = prod[0..5] / fac[0..2]
 *
 * implementation relies on result not being longer than 32 bits
 *
 */

.global mydiv
.func mydiv

prod6=          22      /* 2 extra bytes remainder */
prod7=          23

mydiv:
                /* save work registers */
                push prod0
                push prod1
                push prod6
                push prod7
                push fac1
                push fac2

                /* load 24 bits from arg2 */
                movw zreg, arg2
                ld fac0, z+
                ld fac1, z+
                ld fac2, z+

                /* load 48 bits from arg1 */
                movw zreg, arg1
                ld prod0, z+
                ld prod1, z+
                ld prod2, z+
                ld prod3, z+
                ld prod4, z+
                ld prod5, z+

                clr prod6
                sub prod7, prod7

                ldi loop, 33

1:              rol prod0
                rol prod1
                rol prod2
                rol prod3
                dec loop
                breq 3f

                rol prod4
                rol prod5
                rol prod6
                rol prod7

                sub prod4, fac0
                sbc prod5, fac1
                sbc prod6, fac2
                sbc prod7, zero
                brcc 2f
                add prod4, fac0
                adc prod5, fac1
                adc prod6, fac2
                adc prod7, zero
                clc
                rjmp 1b
2:              sec
                rjmp 1b

                /* store back result */
3:              movw zreg, arg1
                st z+, prod0
                st z+, prod1
                st z+, prod2
                st z+, prod3
                st z+, zero
                st z+, zero

                /* restore work registers */
                pop fac2
                pop fac1
                pop prod7
                pop prod6
                pop prod1
                pop prod0

                ret
.endfunc



/* void mybin2bcd(void *buf)
 *
 * buf[0..3] = bcd(buf[0..3])
 *
 * relies on the input not being bigger than 99999999 = 0x5f5e0ff
 *
 */

.global mybin2bcd
.func mybin2bcd

bcd0=   	16      /* 4 byte bcd result */
bcd1=   	17
bcd2=   	18
bcd3=   	19

bin0=   	20      /* 4 byte binary argument */
bin1=   	21
bin2=   	22
bin3=           23

tmp=    	27

mybin2bcd:
                /* save work registers */
                push bcd0
                push bcd1

                /* load 32 bits from arg1 */
                movw zreg, arg1
                ld bin0, z+
                ld bin1, z+
                ld bin2, z+
                ld bin3, z+

                /* initialize result */
                clr bcd0
                clr bcd1
                clr bcd2
                clr bcd3

                clr zreg+1
                ldi loop, 32

1:              lsl bin0
		rol bin1
		rol bin2
		rol bin3
		rol bcd0
		rol bcd1
		rol bcd2
		rol bcd3
                dec loop
                breq 3f

                /* decimal correction "Dreierkorrektur" */
                ldi zreg, bcd3+1
2:              ld tmp, -z
                subi tmp, -0x03
                sbrc tmp, 3
                st z, tmp
                ld tmp, z
                subi tmp, -0x30
                sbrc tmp, 7
                st z, tmp
                cpi zreg, bcd0
                brne 2b
                rjmp 1b

                /* store back bcd result */
3:              movw zreg, arg1
                st z+, bcd0
                st z+, bcd1
                st z+, bcd2
                st z+, bcd3

                /* restore work registers */
                pop bcd1
                pop bcd0

                ret
.endfunc



/* void mymulby10(void *buf)
 *
 * buf[0..5] = buf[0..5] * 10
 *
 * relies on the input being <= 0x199999
 * (0x199999 * 10 = 0xfffffa)
 *
 */

.global mymulby10
.func mymulby10

akku0=          18
akku1=          19
akku2=          20
akku3=          21
akku4=          22
akku5=          23

mymulby10:
                /* load 48 bits from arg1 */
                movw zreg, arg1
                ld akku0, z+
                ld akku1, z+
                ld akku2, z+
                ld akku3, z+
                ld akku4, z+
                ld akku5, z+

                /* shift left by 1 */
                lsl akku0
                rol akku1
                rol akku2
                rol akku3
                rol akku4
                rol akku5

                /* store back */
                movw zreg, arg1
                st z+, akku0
                st z+, akku1
                st z+, akku2
                st z+, akku3
                st z+, akku4
                st z+, akku5

                /* shift left another 2 bits */
                ldi loop, 2
1:              lsl akku0
                rol akku1
                rol akku2
                rol akku3
                rol akku4
                rol akku5
                dec loop
                brne 1b

        	/* add in single shifted */
                movw zreg, arg1
                ld loop, z
                add akku0, loop
                st z+, akku0
                ld loop, z
                adc akku1, loop
                st z+, akku1
                ld loop, z
                adc akku2, loop
                st z+, akku2
                ld loop, z
                adc akku3, loop
                st z+, akku3
                ld loop, z
                adc akku4, loop
                st z+, akku4
                ld loop, z
                adc akku5, loop
                st z+, akku5

                ret
.endfunc

